from __future__ import absolute_import

from functools import partial

from graphene import relay
from graphene.relay.connection import PageInfo
from graphql_relay.connection.connectiontypes import Edge

from graphene_pynamodb.utils import get_key_name


class PynamoConnectionField(relay.ConnectionField):
    def __init__(self, type, *args, **kwargs):
        super(PynamoConnectionField, self).__init__(
            type,
            *args,
            **kwargs
        )

    @property
    def model(self):
        return self.type._meta.node._meta.model

    @classmethod
    def get_query(cls, model, context, info, args):
        return model.scan

    @classmethod
    def connection_resolver(cls, resolver, connection, model, root, args, context, info):
        iterable = resolver(root, args, context, info)

        first = args.get('first')
        after = args.get('after')
        last = args.get('last')
        before = args.get('before')
        has_previous_page = bool(after)
        page_size = first if first else last if last else None
        connection_type = connection
        pageinfo_type = PageInfo

        # get a full scan query since we have no resolved iterable from relationship or resolver function
        if not iterable and not root:
            query = cls.get_query(model, context, info, args)
            iterable = query()
            if first or last or after or before:
                raise NotImplementedError(
                    "DynamoDB scan operations have no predictable sort. Arguments first, last, after " +
                    "and before will have unpredictable results")

        iterable = iterable if isinstance(iterable, list) else list(iterable) if iterable else []
        if last:
            iterable = iterable[-last:]

        (has_next, edges) = cls.get_edges_from_iterable(iterable, model, edge_type=connection.Edge, after=after,
                                                        page_size=page_size)

        key_name = get_key_name(model)
        try:
            start_cursor = getattr(edges[0].node, key_name)
            end_cursor = getattr(edges[-1].node, key_name)
        except IndexError:
            start_cursor = None
            end_cursor = None

        # Construct the connection
        return connection_type(
            edges=edges,
            page_info=pageinfo_type(
                start_cursor=start_cursor if start_cursor else '',
                end_cursor=end_cursor if end_cursor else '',
                has_previous_page=has_previous_page,
                has_next_page=has_next
            )
        )

    def get_resolver(self, parent_resolver):
        return partial(self.connection_resolver, parent_resolver, self.type, self.model)

    @classmethod
    def get_edges_from_iterable(cls, iterable, model, edge_type=Edge, after=None, page_size=None):
        edges = []
        count = 0
        has_next = False

        for entity in iterable:
            if after:
                if after != str(getattr(entity, get_key_name(model))):
                    continue
                else:
                    after = False
                    continue
            if page_size and count >= page_size:
                has_next = True
                break
            edges.append(edge_type(node=entity, cursor=entity))
            count += 1

        return [has_next, edges]



from py.test import raises

from ..registry import Registry
from ..types import PynamoObjectType
from .models import Reporter


def test_should_raise_if_no_model():
    with raises(Exception) as excinfo:
        class Character1(PynamoObjectType):
            pass
    assert 'valid PynamoDB Model' in str(excinfo.value)


def test_should_raise_if_model_is_invalid():
    with raises(Exception) as excinfo:
        class Character2(PynamoObjectType):

            class Meta:
                model = 1
    assert 'valid PynamoDB Model' in str(excinfo.value)


def test_should_map_fields_correctly():
    class ReporterType2(PynamoObjectType):

        class Meta:
            model = Reporter
            registry = Registry()

    expected_keys = [
        'articles',
        'first_name',
        'last_name',
        'email',
        'pets',
        'id',
        'favorite_article']
    assert all(item in expected_keys for item in ReporterType2._meta.fields.keys())


def test_should_map_only_few_fields():
    class Reporter2(PynamoObjectType):

        class Meta:
            model = Reporter
            only_fields = ('id', 'email')

    assert list(Reporter2._meta.fields.keys()) == ['email', 'id']



from graphene import Field, GlobalID, Int, Interface, ObjectType
from graphene.relay import Node, is_node
from mock import patch

from .models import Article, Reporter
from ..registry import Registry
from ..types import PynamoObjectType

registry = Registry()


class Character(PynamoObjectType):
    '''Character description'''

    class Meta:
        model = Reporter
        registry = registry


class Human(PynamoObjectType):
    '''Human description'''

    pub_date = Int()

    class Meta:
        model = Article
        exclude_fields = ('id',)
        registry = registry
        interfaces = (Node,)


def test_pynamo_interface():
    assert issubclass(Node, Interface)
    assert issubclass(Node, Node)


@patch('graphene_pynamodb.tests.models.Article.get', return_value=Article(id=1))
def test_pynamo_get_node(get):
    human = Human.get_node(1, None, None)
    get.assert_called_with(1)
    assert human.id == 1


def test_objecttype_registered():
    assert issubclass(Character, ObjectType)
    assert Character._meta.model == Reporter
    assert list(Character._meta.fields.keys()) == [
        'articles',
        'email',
        'favorite_article',
        'first_name',
        'id',
        'last_name',
        'pets']


def test_node_idfield():
    idfield = Human._meta.fields['id']
    assert isinstance(idfield, GlobalID)


def test_node_replacedfield():
    idfield = Human._meta.fields['pub_date']
    assert isinstance(idfield, Field)
    assert idfield.type == Int


def test_object_type():
    assert issubclass(Human, ObjectType)
    assert list(Human._meta.fields.keys()) == ['id', 'headline', 'reporter', 'pub_date']
    assert is_node(Human)



class Registry(object):
    def __init__(self):
        self._registry = {}
        self._registry_models = {}
        self._registry_composites = {}

    def register(self, cls):
        from .types import PynamoObjectType
        assert issubclass(
            cls, PynamoObjectType), 'Only PynamoObjectType can be registered, received "{}"'.format(
            cls.__name__)
        assert cls._meta.registry == self, 'Registry for a Model have to match.'
        # assert self.get_type_for_model(cls._meta.model) in [None, cls], (
        #     'SQLAlchemy model "{}" already associated with '
        #     'another type "{}".'
        # ).format(cls._meta.model, self._registry[cls._meta.model])
        self._registry[cls._meta.model] = cls

    def get_type_for_model(self, model):
        return self._registry.get(model)


registry = None


def get_global_registry():
    global registry
    if not registry:
        registry = Registry()
    return registry



import graphene
from graphene import Dynamic
from graphene import Node
from graphene.types.json import JSONString
from py.test import raises
from pynamodb.attributes import (BinaryAttribute, BinarySetAttribute,
                                 BooleanAttribute, JSONAttribute,
                                 NumberAttribute, NumberSetAttribute,
                                 UnicodeAttribute, UnicodeSetAttribute,
                                 UTCDateTimeAttribute)
from pynamodb.models import Model

from graphene_pynamodb.relationships import OneToMany
from .models import Article, Reporter
from .. import PynamoConnectionField
from .. import PynamoObjectType
from ..converter import convert_pynamo_attribute


def assert_attribute_conversion(attribute, graphene_field, **kwargs):
    graphene_type = convert_pynamo_attribute(attribute, attribute)
    assert isinstance(graphene_type, graphene_field)
    field = graphene_type.Field()
    return field


def test_should_unknown_pynamo_field_raise_exception():
    with raises(Exception) as excinfo:
        convert_pynamo_attribute(None, None, None)
    assert 'Don\'t know how to convert the PynamoDB attribute' in str(excinfo.value)


def test_should_datetime_convert_string():
    assert_attribute_conversion(UTCDateTimeAttribute(), graphene.String)


def test_should_string_convert_string():
    assert_attribute_conversion(UnicodeAttribute(), graphene.String)


def test_should_binary_convert_string():
    assert_attribute_conversion(BinaryAttribute(), graphene.String)


def test_should_number_convert_int():
    assert_attribute_conversion(NumberAttribute(), graphene.Int)


def test_should_boolean_convert_boolean():
    assert_attribute_conversion(BooleanAttribute(), graphene.Boolean)


def test_should_string_set_convert_list():
    assert_attribute_conversion(UnicodeSetAttribute(), graphene.List)


def test_should_number_set_convert_list():
    assert_attribute_conversion(NumberSetAttribute(), graphene.List)


def test_should_binary_set_convert_list():
    assert_attribute_conversion(BinarySetAttribute(), graphene.List)


def test_should_jsontype_convert_jsonstring():
    assert_attribute_conversion(JSONAttribute(), JSONString)


def test_should_onetoone_convert_field():
    class A(PynamoObjectType):
        class Meta:
            model = Article
            interfaces = (Node,)

    dynamic_field = convert_pynamo_attribute(Reporter.favorite_article, Reporter.favorite_article, A._meta.registry)
    assert isinstance(dynamic_field, Dynamic)
    graphene_type = dynamic_field.get_type()
    assert isinstance(graphene_type, graphene.Field)
    assert graphene_type.type == A


def test_should_onetomany_convert_nonnode_field():
    class A(PynamoObjectType):
        class Meta:
            model = Article

    dynamic_field = convert_pynamo_attribute(Reporter.articles, Reporter.articles, A._meta.registry)
    assert isinstance(dynamic_field, Dynamic)
    graphene_type = dynamic_field.get_type()
    assert isinstance(graphene_type, graphene.Field)
    assert graphene_type.type == graphene.List(A)


def test_should_onetomany_none_for_unknown_type():
    class ModelA(Model):
        pass

    class ModelB(Model):
        a = OneToMany(ModelA)

    class A(PynamoObjectType):
        class Meta:
            model = ModelB

    dynamic_field = convert_pynamo_attribute(ModelB.a, ModelB.a, A._meta.registry)
    assert isinstance(dynamic_field, Dynamic)
    assert dynamic_field.get_type() is None


def test_should_onetomany_convert_field():
    class A(PynamoObjectType):
        class Meta:
            model = Article
            interfaces = (Node,)

    dynamic_field = convert_pynamo_attribute(Reporter.articles, Reporter.articles, A._meta.registry)
    assert isinstance(dynamic_field, Dynamic)
    graphene_type = dynamic_field.get_type()
    assert isinstance(graphene_type, PynamoConnectionField)



import graphene
import pytest
from graphene import Node
from mock import MagicMock
from wrapt import ObjectProxy

from .models import Reporter, Article
from ..relationships import OneToOne, OneToMany, RelationshipResult
from ..types import PynamoObjectType


def setup_fixtures():
    class ReporterType(PynamoObjectType):
        class Meta:
            model = Reporter
            interfaces = (Node,)

    class ArticleType(PynamoObjectType):
        class Meta:
            model = Article
            interfaces = (Node,)

    reporter1 = Reporter(1, first_name="John", last_name="Snow")
    article1 = Article(1, headline="Hi!", reporter=reporter1)
    article2 = Article(2, headline="Lame Article", reporter=reporter1)
    reporter1.articles = [article1, article2]

    return {
        'ReporterType': ReporterType,
        'ArticleType': ArticleType,
        'reporter1': reporter1,
        'article1': article1,
        'article2': article2
    }


def test_relationships_should_resolve_well():
    fixtures = setup_fixtures()

    class Query(graphene.ObjectType):
        reporter = graphene.Field(fixtures["ReporterType"])
        articles = graphene.List(fixtures["ArticleType"])

        def resolve_reporter(self, *args, **kwargs):
            return fixtures["reporter1"]

        def resolve_articles(self, *args, **kwargs):
            return fixtures["reporter1"].articles

    query = '''
      query AuthorQuery {
        reporter {
          firstName,
          lastName,
          articles {
            edges {
              node {
                id
              }
            }
          }
        }
        articles {
          headline
          reporter {
            id
          }
        }
      }
    '''
    expected = {
        'reporter': {
            'firstName': 'John',
            'lastName': 'Snow',
            'articles': {
                'edges': [{
                    'node': {
                        'id': 'QXJ0aWNsZVR5cGU6MQ==',
                    }
                }, {
                    'node': {
                        'id': 'QXJ0aWNsZVR5cGU6Mg=='
                    }
                }]
            }
        },
        'articles': [{
            'headline': 'Hi!',
            'reporter': {
                'id': 'UmVwb3J0ZXJUeXBlOjE='
            }
        }, {
            'headline': 'Lame Article',
            'reporter': {
                'id': 'UmVwb3J0ZXJUeXBlOjE='
            }
        }]
    }

    schema = graphene.Schema(query=Query)

    result = schema.execute(query)
    assert not result.errors
    assert result.data['reporter'] == expected['reporter']
    assert all(item in result.data['articles'] for item in expected['articles'])


def test_relationships_should_raiseerror():
    with pytest.raises(TypeError):
        OneToOne(object)
    with pytest.raises(TypeError):
        OneToMany(object)


def test_onetoone_should_serialize_well():
    fixtures = setup_fixtures()
    relationship = OneToOne(Article)
    assert relationship.serialize(fixtures["article1"]) == "1"


def test_onetoone_should_deserialize_well():
    fixtures = setup_fixtures()
    relationship = OneToOne(Article)
    article = relationship.deserialize(1)
    assert isinstance(article, Article.__class__)
    assert article.id == fixtures["article1"].id
    assert article.headline == fixtures["article1"].headline


def test_onetomany_should_serialize_well():
    fixtures = setup_fixtures()
    relationship = OneToMany(Article)
    assert relationship.serialize([fixtures["article1"], fixtures["article2"]]) == ['1', '2']


def test_onetomany_should_deserialize_well():
    relationship = OneToMany(Article)

    articles = relationship.deserialize([1, 3])
    assert len(articles) == 2
    # test before db call (lazy)
    assert articles[0] == Article(1)
    assert articles[1] == Article(3)
    # test db call
    assert articles[0].headline == "Hi!"
    assert articles[1].headline == "My Article"


def test_result_should_be_lazy():
    MockArticle = ObjectProxy(Article)
    MockArticle.get = MagicMock(return_value=Article.get(1))
    relationship = RelationshipResult('id', 1, MockArticle)

    # test before db call (lazy)
    assert relationship.id == 1
    MockArticle.get.assert_not_called()
    # test db call only once
    assert relationship.headline == "Hi!"
    assert relationship.reporter.id == 1
    assert relationship.headline == "Hi!"
    MockArticle.get.assert_called_once_with(1)


def test_relationships_should_compare_well():
    article1 = Article(1, healine="test")
    article2 = Article(2, healine="test")
    rel1 = RelationshipResult('id', 1, Article)
    assert rel1 == article1
    assert rel1 != article2


def test_result_should_check_type():
    with pytest.raises(TypeError):
        RelationshipResult('id', 1, PynamoObjectType)


def test_onetoone_should_handle_not_being_lazy():
    MockArticle = ObjectProxy(Article)
    MockArticle.get = MagicMock(return_value=Article.get(1))
    relationship = OneToOne(MockArticle, lazy=False)
    article = relationship.deserialize(1)
    MockArticle.get.assert_called_once_with(1)

    # Access fields that would normally trigger laxy loading
    assert article.id == 1
    assert article.headline == "Hi!"
    assert article.reporter.id == 1
    assert article.headline == "Hi!"
    # make sure our call count is still 1
    MockArticle.get.assert_called_once_with(1)


def test_onetomany_should_handle_not_being_lazy():
    MockArticle = ObjectProxy(Article)
    MockArticle.get = MagicMock()
    MockArticle.batch_get = MagicMock(return_value=Article.batch_get([1, 3]))
    relationship = OneToMany(MockArticle, lazy=False)
    articles = list(relationship.deserialize([1, 3]))
    MockArticle.batch_get.assert_called_once()
    MockArticle.get.assert_not_called()
    # Access fields that would normally trigger laxy loading
    # order is not guaranteed in batch_get
    if articles[0].id == 1:
        assert articles[0].headline == "Hi!"
        assert articles[1].id == 3
        assert articles[1].headline == "My Article"
    else:
        assert articles[0].id == 3
        assert articles[0].headline == "My Article"
        assert articles[1].id == 1
        assert articles[1].headline == "Hi!"
    # make sure our call count is still 1
    MockArticle.batch_get.assert_called_once()
    MockArticle.get.assert_not_called()



import graphene
from graphene.relay import Node

from .models import Article, Editor, Reporter
from ..fields import PynamoConnectionField
from ..types import PynamoObjectType


def setup_fixtures():
    if not Reporter.exists():
        Reporter.create_table(read_capacity_units=1, write_capacity_units=1, wait=True)
        reporter = Reporter(id=1, first_name='ABA', last_name='X')
        reporter.articles = [Article(1), Article(3)]
        reporter.save()
        reporter2 = Reporter(id=2, first_name='ABO', last_name='Y')
        reporter2.save()
    if not Article.exists():
        Article.create_table(read_capacity_units=1, write_capacity_units=1, wait=True)
        article1 = Article(id=1, headline='Hi!', reporter=Reporter(1))
        article1.save()
        article2 = Article(id=3, headline='My Article', reporter=Reporter(1))
        article2.save()
    if not Editor.exists():
        Editor.create_table(read_capacity_units=1, write_capacity_units=1, wait=True)
        editor = Editor(id='1', name='John')
        editor.save()


setup_fixtures()


def test_should_query_well():
    class ReporterType(PynamoObjectType):
        class Meta:
            model = Reporter

    class Query(graphene.ObjectType):
        reporter = graphene.Field(ReporterType)
        reporters = graphene.List(ReporterType)

        def resolve_reporter(self, *args, **kwargs):
            return Reporter.get(1)

        def resolve_reporters(self, *args, **kwargs):
            return Reporter.scan()

    query = '''
        query ReporterQuery {
          reporter {
            firstName,
            lastName,
            email
          }
          reporters {
            firstName
          }
        }
    '''
    expected = {
        'reporter': {
            'email': None,
            'firstName': 'ABA',
            'lastName': 'X'
        },
        'reporters': [{
            'firstName': 'ABO',
        }, {
            'firstName': 'ABA',
        }]
    }
    schema = graphene.Schema(query=Query)
    result = schema.execute(query)
    assert not result.errors
    assert result.data['reporter'] == expected['reporter']
    assert all(item in result.data['reporters'] for item in expected['reporters'])


def test_should_node():
    class ReporterNode(PynamoObjectType):
        class Meta:
            model = Reporter
            interfaces = (Node,)

        @classmethod
        def get_node(cls, id, context, info):
            return Reporter(id=2, first_name='Cookie Monster')

    class ArticleNode(PynamoObjectType):
        class Meta:
            model = Article
            interfaces = (Node,)

        @classmethod
        def get_node(cls, id, context, info):
            return Article(id=1, headline='Article node')

    class Query(graphene.ObjectType):
        node = Node.Field()
        reporter = graphene.Field(ReporterNode)
        article = graphene.Field(ArticleNode)
        all_articles = PynamoConnectionField(ArticleNode)

        def resolve_reporter(self, *args, **kwargs):
            return Reporter.get(1)

        def resolve_article(self, *args, **kwargs):
            return Article.get(1)

    query = '''
        query ReporterQuery {
          reporter {
            id,
            firstName,
            articles {
              edges {
                node {
                  headline
                }
              }
            }
            lastName,
            email
          }
          allArticles {
            edges {
              node {
                headline
              }
            }
          }
          myArticle: node(id:"QXJ0aWNsZU5vZGU6MQ==") {
            id
            ... on ReporterNode {
                firstName
            }
            ... on ArticleNode {
                headline
            }
          }
        }
    '''
    expected = {
        'reporter': {
            'id': 'UmVwb3J0ZXJOb2RlOjE=',
            'firstName': 'ABA',
            'lastName': 'X',
            'email': None,
            'articles': {
                'edges': [
                    {
                        'node': {
                            'headline': 'Hi!'
                        }
                    },
                    {
                        'node': {
                            'headline': 'My Article'
                        }
                    }
                ]
            }
        },
        'allArticles': {
            'edges': [{
                'node': {
                    'headline': 'Hi!'
                }
            }]
        },
        'myArticle': {
            'id': 'QXJ0aWNsZU5vZGU6MQ==',
            'headline': 'Article node'
        }
    }
    schema = graphene.Schema(query=Query)
    result = schema.execute(query)
    assert not result.errors
    assert all(item in expected['reporter'] for item in result.data['reporter'])
    assert all(item in expected['reporter']['articles'] for item in result.data['reporter']['articles'])
    assert result.data['myArticle'] == expected['myArticle']
    assert all(item in result.data['allArticles'] for item in expected['allArticles'])


def test_should_custom_identifier():
    class EditorNode(PynamoObjectType):
        class Meta:
            model = Editor
            interfaces = (Node,)

    class Query(graphene.ObjectType):
        node = Node.Field()
        all_editors = PynamoConnectionField(EditorNode)

    query = '''
        query EditorQuery {
          allEditors {
            edges {
                node {
                    id,
                    name
                }
            }
          },
          node(id: "RWRpdG9yTm9kZTox") {
            ...on EditorNode {
              name
            }
          }
        }
    '''
    expected = {
        'allEditors': {
            'edges': [{
                'node': {
                    'id': 'RWRpdG9yTm9kZTox',
                    'name': 'John'
                }
            }]
        },
        'node': {
            'name': 'John'
        }
    }

    schema = graphene.Schema(query=Query)
    result = schema.execute(query)
    assert not result.errors
    assert result.data['allEditors'] == expected['allEditors']


def test_should_mutate_well():
    class EditorNode(PynamoObjectType):
        class Meta:
            model = Editor
            interfaces = (Node,)

    class ReporterNode(PynamoObjectType):
        class Meta:
            model = Reporter
            interfaces = (Node,)

        @classmethod
        def get_node(cls, id, info):
            return Reporter(id=2, first_name='Cookie Monster')

    class ArticleNode(PynamoObjectType):
        class Meta:
            model = Article
            interfaces = (Node,)

    class CreateArticle(graphene.Mutation):
        class Input:
            headline = graphene.String()
            reporter_id = graphene.ID()

        ok = graphene.Boolean()
        article = graphene.Field(ArticleNode)

        @classmethod
        def mutate(cls, instance, args, context, info):
            new_article = Article(
                id=3,
                headline=args.get('headline'),
                reporter=Reporter.get(int(args.get('reporter_id'))),
            )
            new_article.save()
            ok = True

            return CreateArticle(article=new_article, ok=ok)

    class Query(graphene.ObjectType):
        node = Node.Field()

    class Mutation(graphene.ObjectType):
        create_article = CreateArticle.Field()

    query = '''
        mutation ArticleCreator {
          createArticle(
            headline: "My Article"
            reporterId: "1"
          ) {
            ok
            article {
              headline
              reporter {
                id
              }
            }
          }
        }
    '''
    expected = {
        'createArticle': {
            'ok': True,
            'article': {
                'headline': 'My Article',
                'reporter': {
                    'id': 1
                }
            }
        },
    }

    schema = graphene.Schema(query=Query, mutation=Mutation)
    result = schema.execute(query)
    assert not result.errors
    assert result.data['createArticle']['ok'] == expected['createArticle']['ok']
    assert all(item in result.data['createArticle']['article'] for item in expected['createArticle']['article'])


def test_should_return_empty_cursors_on_empty():
    class ArticleNode(PynamoObjectType):
        class Meta:
            model = Article
            interfaces = (Node,)

    class ReporterNode(PynamoObjectType):
        class Meta:
            model = Reporter
            interfaces = (Node,)

    class Query(graphene.ObjectType):
        node = Node.Field()
        reporter = graphene.Field(ReporterNode)

        def resolve_reporter(self, *args, **kwargs):
            return Reporter.get(2)

    query = '''
        query ReporterQuery {
          reporter {
            id,
            firstName,
            articles(first: 1) {
              edges {
                node {
                  id
                  headline
                }
              }
              pageInfo {
                hasNextPage
                hasPreviousPage
                startCursor
                endCursor
              }
            }
            lastName,
            email
          }
        }
    '''
    expected = {
        'reporter': {
            'articles': {
                'edges': [],
                'pageInfo': {
                    'hasNextPage': False,
                    'hasPreviousPage': False,
                    'startCursor': '',
                    'endCursor': ''
                }
            }
        }
    }

    schema = graphene.Schema(query=Query)
    result = schema.execute(query)
    assert not result.errors
    assert result.data['reporter']['articles']['edges'] == expected['reporter']['articles']['edges']
    assert result.data['reporter']['articles']['pageInfo'] == expected['reporter']['articles']['pageInfo']


def test_should_support_first():
    class ArticleNode(PynamoObjectType):
        class Meta:
            model = Article
            interfaces = (Node,)

    class ReporterNode(PynamoObjectType):
        class Meta:
            model = Reporter
            interfaces = (Node,)

    class Query(graphene.ObjectType):
        node = Node.Field()
        reporter = graphene.Field(ReporterNode)

        def resolve_reporter(self, *args, **kwargs):
            return Reporter.get(1)

    query = '''
        query ReporterQuery {
          reporter {
            id,
            firstName,
            articles(first: 1) {
              edges {
                node {
                  id
                  headline
                }
              }
              pageInfo {
                hasNextPage
                hasPreviousPage
                startCursor
                endCursor
              }
            }
            lastName,
            email
          }
        }
    '''
    expected = {
        'reporter': {
            'articles': {
                'edges': [{
                    'node': {
                        'id': 'QXJ0aWNsZU5vZGU6MQ==',
                        'headline': 'Hi!'
                    }
                }],
                'pageInfo': {
                    'hasNextPage': True,
                    'hasPreviousPage': False,
                    'startCursor': '1',
                    'endCursor': '1'
                }
            }
        }
    }

    schema = graphene.Schema(query=Query)
    result = schema.execute(query)
    assert not result.errors
    assert result.data['reporter']['articles']['edges'] == expected['reporter']['articles']['edges']
    assert result.data['reporter']['articles']['pageInfo'] == expected['reporter']['articles']['pageInfo']


def test_should_support_last():
    class ArticleNode(PynamoObjectType):
        class Meta:
            model = Article
            interfaces = (Node,)

    class ReporterNode(PynamoObjectType):
        class Meta:
            model = Reporter
            interfaces = (Node,)

    class Query(graphene.ObjectType):
        node = Node.Field()
        reporter = graphene.Field(ReporterNode)

        def resolve_reporter(self, *args, **kwargs):
            return Reporter.get(1)

    query = '''
        query ReporterQuery {
          reporter {
            id,
            firstName,
            articles(last: 1) {
              edges {
                node {
                  id
                  headline
                }
              }
            }
            lastName,
            email
          }
        }
    '''
    expected = {
        'reporter': {
            'articles': {
                'edges': [{
                    'node': {
                        'id': 'QXJ0aWNsZU5vZGU6Mw==',
                        'headline': 'My Article'
                    }
                }]
            }
        }
    }

    schema = graphene.Schema(query=Query)
    result = schema.execute(query)
    assert not result.errors
    assert result.data['reporter']['articles']['edges'] == expected['reporter']['articles']['edges']


def test_should_support_after():
    class ArticleNode(PynamoObjectType):
        class Meta:
            model = Article
            interfaces = (Node,)

    class ReporterNode(PynamoObjectType):
        class Meta:
            model = Reporter
            interfaces = (Node,)

    class Query(graphene.ObjectType):
        node = Node.Field()
        reporter = graphene.Field(ReporterNode)

        def resolve_reporter(self, *args, **kwargs):
            return Reporter.get(1)

    query = '''
        query ReporterQuery {
          reporter {
            id,
            firstName,
            articles(after: "1") {
              edges {
                node {
                  id
                  headline
                }
              }
            }
            lastName,
            email
          }
        }
    '''
    expected = {
        'reporter': {
            'articles': {
                'edges': [{
                    'node': {
                        'id': 'QXJ0aWNsZU5vZGU6Mw==',
                        'headline': 'My Article'
                    }
                }]
            }
        }
    }

    schema = graphene.Schema(query=Query)
    result = schema.execute(query)
    assert not result.errors
    assert result.data['reporter']['articles']['edges'] == expected['reporter']['articles']['edges']


def test_root_scan_should_warn_on_params():
    class ArticleNode(PynamoObjectType):
        class Meta:
            model = Article
            interfaces = (Node,)

    class Query(graphene.ObjectType):
        node = Node.Field()
        articles = PynamoConnectionField(ArticleNode)

    query = '''
        query ArticlesQuery {
          articles(after: "1") {
            edges {
              node {
                id
                headline
              }
            }
          }
        }
    '''

    schema = graphene.Schema(query=Query)
    result = schema.execute(query)
    assert result.errors
    assert isinstance(result.errors[0].original_error, NotImplementedError)


def test_root_scan_should_fastforward_on_after():
    class ArticleNode(PynamoObjectType):
        class Meta:
            model = Article
            interfaces = (Node,)

    class Query(graphene.ObjectType):
        node = Node.Field()
        articles = PynamoConnectionField(ArticleNode)

        def resolve_articles(self, *args, **kwargs):
            return [
                Article(1, headline='One'),
                Article(2, headline='Two'),
                Article(3, headline='Three'),
                Article(4, headline='Four')
            ]

    query = '''
        query ArticlesQuery {
          articles(after: "2", first: 1) {
            edges {
              node {
                id
                headline
              }
            }
          }
        }
    '''
    expected = [{
        'node': {
            'headline': 'Three',
            'id': 'QXJ0aWNsZU5vZGU6Mw=='
        }
    }]

    schema = graphene.Schema(query=Query)
    result = schema.execute(query)
    assert not result.errors
    assert result.data['articles']['edges'] == expected
